# Copyright © 2019 Lars Pastewka
#
# µSpectre is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Lesser Public License as
# published by the Free Software Foundation, either version 3, or (at
# your option) any later version.
#
# µSpectre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with µSpectre; see the file COPYING. If not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this Program, or any covered work, by linking or combining it
# with proprietary FFT implementations or numerical libraries, containing parts
# covered by the terms of those libraries' licenses, the licensors of this
# Program grant you additional permission to convey the resulting work.

#
# This is the most minimal-idiotic way of discovering the version that I
# could come up with. It deals with the following issues:
# * If we are installed, we can get the version from package metadata,
#   either via importlib.metadata or from pkg_resources. This also holds for
#   wheels that contain the metadata. We are good! Yay!
# * If we are not installed, there are two options:
#   - We are working within the source git repository. Then
#        git describe --tags --always
#     yields a reasonable version descriptor, but that is unfortunately not
#     PEP 440 compliant (see https://peps.python.org/pep-0440/). We need to
#     mangle the version string to yield something compatible.
# - If we are not in a source directory, check if muGrid's version.cc has
#   been created and extract version from that file.
# - If we install from a source tarball, all version information may be lost.
#   Fortunately, Meson uses git archive to create the source tarball, which
#   replaces certain tags with commit information. Unfortunately, what this
#   yields is different from `git describe` - in particular, it only yields the
#   tag (which contains the version information) if we are *exactly* on the
#   tag commit. (`git describe` tells us the distance from the latest tag.) We
#   need to extract the version information from the string provided, but if
#   we are not on the tag we can only return a bogus version (here 0.0.0.0).
#   It works for releases, but not for a tarball generated from a random
#   commit. I am not happy and open for suggestions.
#

import re
import subprocess
import sys

class CannotDiscoverVersion(Exception):
    pass


def get_version_from_pkg_info():
    """
    Discover version from PKG-INFO file. This is never dirty and we cannot
    obtain the git hash. This is typically no problem since the version
    uniquely identifies the commit.
    """
    try:
        with open('PKG-INFO', 'r') as f:
            l = f.readline()
            while l:
                if l.startswith('Version:'):
                    return False, l[8:].strip(), '(PKG-INFO)'
                l = f.readline()
    except FileNotFoundError:
        raise CannotDiscoverVersion("File 'PKG-INFO' does not exist.")
    raise CannotDiscoverVersion("No line starting with 'Version:' in 'PKG-INFO'.")


def get_version_from_git():
    """
    Discover muSpectre version from git repository. We get all information,
    including whether the version is dirty and the git hash.
    """
    git_describe = subprocess.run(
        ['git', 'describe', '--tags', '--dirty', '--always'],
        stdout=subprocess.PIPE)
    if git_describe.returncode != 0:
        raise CannotDiscoverVersion('git execution failed')
    version = git_describe.stdout.decode('latin-1').strip()
    git_show = subprocess.run(
        ['git', 'show', '-s', '--format=%H'],
        stdout=subprocess.PIPE)
    if git_show.returncode != 0:
        raise CannotDiscoverVersion('git execution failed')
    git_hash = git_show.stdout.decode('latin-1').strip()

    dirty = version.endswith('-dirty')

    # Make version PEP 440 compliant
    if dirty:
        version = version.replace('-dirty', '')
    version = version.replace('-', '+', 1)
    if dirty:
        version += '-dirty'

    return dirty, version, git_hash


def get_version_from_cc(fn):
    """
    Extract version from a version.cc file (that is typically created by the)
    build system. This will also yield dirty state and git hash, if present
    in the cc file.
    """
    text = open(fn, 'r').read()
    dirty = bool(re.search('constexpr bool git_dirty{(true|false)};',
                           text).group(1))
    version = re.search(
        r'constexpr char git_describe[]{"([A-Za-z0-9_.+-]*)"};', text
    ).group(1)
    git_hash = re.search(r'constexpr char git_hash[]{"([A-Za-z0-9_.-]*)"};',
                         text).group(1)

    return dirty, version, git_hash


#
# Discover version and write version.cc if version comes from git
#

try:
    try:
        # First try is to look inside PKG-INFO file.
        dirty, version, git_hash = get_version_from_pkg_info()
    except CannotDiscoverVersion:
        dirty, version, git_hash = get_version_from_git()
except CannotDiscoverVersion:
    # Detection via PKG-INFO and git failed. Get version from version.cc file.
    dirty, version, git_hash = \
        get_version_from_cc('src/libmugrid/version.cc')

#
# Print version to screen
#

if len(sys.argv) > 1 and '--full' in sys.argv:
    # True and false need to be C++ compatible
    print('true' if dirty else 'false', version, git_hash)
else:
    print(version)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file   python_binding_tests.py

@author Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date   16 Oct 2019

@brief  Unit tests for python bindings

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import unittest

import muGrid

from python_common_tests import *

from python_communicator_tests import *
from python_field_tests import *

if 'FileIONetCDF' in dir(muGrid):
    from python_file_io_test import FileIOTest

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file     python_field_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    18 Oct 2019

@brief   test field collections and fields

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import unittest
import numpy as np

import muGrid


class FieldCheck(unittest.TestCase):
    def setUp(self):
        self.nb_grid_pts = (10, 11, 21)

    def test_buffer_size_one_quad_pt(self):
        values = np.random.rand(*self.nb_grid_pts)

        nb_grid_pts2 = (4,) + self.nb_grid_pts
        values2 = np.random.rand(*nb_grid_pts2)

        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts)

        # Scalar
        fs = fc.register_real_field("test-field-scalar")
        self.assertEqual(fs.array(muGrid.Pixel).shape, self.nb_grid_pts)
        self.assertEqual(fs.array(muGrid.SubPt).shape, (1,) + self.nb_grid_pts)
        self.assertEqual(fs.p.shape, self.nb_grid_pts)
        self.assertEqual(fs.s.shape, (1,) + self.nb_grid_pts)

        # Single component
        f = fc.register_real_field("test-field", 1)
        self.assertEqual(f.array(muGrid.Pixel).shape, (1,) + self.nb_grid_pts)
        self.assertEqual(f.array(muGrid.SubPt).shape, (1, 1) + self.nb_grid_pts)
        self.assertEqual(f.p.shape, (1,) + self.nb_grid_pts)
        self.assertEqual(f.s.shape, (1, 1) + self.nb_grid_pts)

        # Four components
        f2 = fc.register_real_field("test-field2", 4)
        self.assertEqual(f2.array(muGrid.Pixel).shape, nb_grid_pts2)
        self.assertEqual(f2.array(muGrid.SubPt).shape, (4, 1) + self.nb_grid_pts)
        self.assertEqual(f2.p.shape, nb_grid_pts2)
        self.assertEqual(f2.s.shape, (4, 1) + self.nb_grid_pts)

        # Check that we get those fields back
        self.assertEqual(fc.get_field('test-field-scalar'), fs)
        self.assertEqual(fc.get_field('test-field'), f)
        self.assertEqual(fc.get_field('test-field2'), f2)

        # Check strides
        self.assertEqual(fs.stride(muGrid.Pixel), 1)
        self.assertEqual(f.stride(muGrid.Pixel), 1)
        self.assertEqual(f2.stride(muGrid.Pixel), 4)
        self.assertEqual(fs.stride(muGrid.SubPt), 1)
        self.assertEqual(f.stride(muGrid.SubPt), 1)
        self.assertEqual(f2.stride(muGrid.SubPt), 4)

        # Check subpoint-shaped convenience access
        self.assertEqual(fs.s.shape, (1,) + self.nb_grid_pts)
        with self.assertRaises(RuntimeError):
            fs.s = np.zeros(self.nb_grid_pts)
        fs.s = values.reshape((1,) + self.nb_grid_pts)
        np.testing.assert_allclose(fs.s, values.reshape((1,) + self.nb_grid_pts))

        self.assertEqual(f.s.shape, (1, 1) + self.nb_grid_pts)
        with self.assertRaises(RuntimeError):
            f.s = np.zeros((3,) + self.nb_grid_pts)
        f.s = values.reshape((1, 1) + self.nb_grid_pts)
        np.testing.assert_allclose(f.s, values.reshape((1, 1) + self.nb_grid_pts))

        self.assertEqual(f2.s.shape, (4, 1) + self.nb_grid_pts)
        with self.assertRaises(RuntimeError):
            # This has shape (4, 1, 10, 11, 21) and should fail
            f2.s = values2

        # Check pixel-shaped convenience access
        self.assertEqual(fs.p.shape, self.nb_grid_pts)
        with self.assertRaises(RuntimeError):
            fs.p = np.zeros((3,) + self.nb_grid_pts)
        fs.p = values
        np.testing.assert_allclose(fs.p, values)

        self.assertEqual(f.p.shape, (1,) + self.nb_grid_pts)
        with self.assertRaises(RuntimeError):
            f.p = np.zeros((3, 1) + self.nb_grid_pts)
        f.p = values.reshape((1,) + self.nb_grid_pts)
        np.testing.assert_allclose(f.p, values.reshape((1,) + self.nb_grid_pts))

        self.assertEqual(f2.p.shape, nb_grid_pts2)
        with self.assertRaises(RuntimeError):
            f2.p = np.zeros((3,) + self.nb_grid_pts)
        f2.p = values2
        np.testing.assert_allclose(f2.p, values2)

    def test_buffer_size_four_quad_pt(self):
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts, sub_pts={'quad': 4})
        # Single component
        f = fc.register_real_field("test-field", 1, 'quad')
        self.assertEqual(f.array(muGrid.Pixel).shape,
                         (4,) + self.nb_grid_pts)
        self.assertEqual(f.array(muGrid.SubPt).shape,
                         (1, 4) + self.nb_grid_pts)
        # Four components
        f2 = fc.register_real_field("test-field2", 3, 'quad')
        self.assertEqual(f2.array(muGrid.Pixel).shape,
                         (4 * 3,) + self.nb_grid_pts)
        self.assertEqual(f2.array(muGrid.SubPt).shape,
                         (3, 4) + self.nb_grid_pts)
        # Check that we get those fields back
        self.assertEqual(fc.get_field('test-field'), f)
        self.assertEqual(fc.get_field('test-field2'), f2)
        # Check strides
        self.assertEqual(f.stride(muGrid.Pixel), 4)
        self.assertEqual(f2.stride(muGrid.Pixel), 3 * 4)
        self.assertEqual(f.stride(muGrid.SubPt), 1)
        self.assertEqual(f2.stride(muGrid.SubPt), 3)

    def test_buffer_access(self):
        fc = muGrid.GlobalFieldCollection(len(self.nb_grid_pts), {'quad': 4})
        fc.initialise(self.nb_grid_pts, self.nb_grid_pts)
        # Single component
        f = fc.register_real_field("test-field", 1, 'quad')
        arr1 = np.array(f, copy=False)
        self.assertFalse(arr1.flags.owndata)
        self.assertTrue(arr1.flags.f_contiguous)
        arr2 = f.array(muGrid.Pixel)
        self.assertFalse(arr2.flags.owndata)
        self.assertTrue(arr2.flags.f_contiguous)

        arr1[0, 0, 2, 1, 0] = 10  # f-contiguous
        self.assertEqual(arr2[0, 2, 1, 0], 10)  # f-contiguous

        f2 = fc.register_real_field("test-field-2", 3, 'quad')
        self.assertTrue(f.p.data != f2.p.data)

    def test_col_major(self):
        dims = (3, 4)
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts, self.nb_grid_pts,
                                          (0,) * len(self.nb_grid_pts),
                                          {})
        f = fc.register_real_field("test-field", dims)
        a = np.array(f)
        self.assertTrue(a.flags.f_contiguous)
        self.assertEqual(a.shape, tuple(list(dims) + [1, ] + list(self.nb_grid_pts)))
        strides = np.append([1], np.cumprod(dims))
        strides = np.append(strides, strides[-1])
        strides = np.append(strides,
                            strides[-1] * np.cumprod(self.nb_grid_pts))
        strides = 8 * strides[:-1]
        self.assertEqual(a.strides, tuple(strides))

    def test_row_major(self):
        # We need to specify the storage order twice and they mean different
        # things.
        dims = (3, 4)
        fc = muGrid.GlobalFieldCollection(
            self.nb_grid_pts, self.nb_grid_pts, [0] * len(self.nb_grid_pts),
            muGrid.StorageOrder.RowMajor, {}, muGrid.StorageOrder.RowMajor)
        f = fc.register_real_field("test-field", dims)
        a = np.array(f)
        self.assertTrue(a.flags.c_contiguous)
        self.assertEqual(a.shape, tuple(list(dims) + [1, ] + list(self.nb_grid_pts)))
        strides = np.append([1], np.cumprod(self.nb_grid_pts[::-1]))
        strides = np.append(strides, strides[-1])
        strides = np.append(strides, strides[-1] * np.cumprod(dims[::-1]))
        strides = 8 * strides[-2::-1]
        self.assertEqual(a.strides, tuple(strides))

    def test_local_field_collection(self):
        lfc_name = "ArbitraryLocalFieldCollectionName"
        lfc_field_name = "MyLocalField"
        fc = muGrid.LocalFieldCollection(3, lfc_name)
        fc.add_pixel(1)  # add pixel with global_index = 1
        fc.initialise()
        field = fc.register_real_field(lfc_field_name, 1, 'pixel')
        field_array = np.array(field, copy=False)
        field_array[:] = 5  # assigne value 5 to pixel 1

        self.assertEqual(fc.get_name(), lfc_name)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file python_muGrid_license_test.py

@author Ali Falsafi<ali.falsafi @epfl.ch>

@date 17 Sep 2019

@brief description

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import unittest
import sys
import os
sys.path.insert(1, os.path.join(sys.path[0], '..'))
import python_license_test as lic_test


# muGrid_sources = ["../src/libmugrid", "../tests/libmugrid"]

lic_paras = [" µGrid is free software; you can redistribute it and/or "
             "modify it under the terms of the GNU Lesser General Public "
             "License as published by the Free Software Foundation, either"
             " version 3, or (at your option) any later version. ",
             " µGrid is distributed in the hope that it will be useful,"
             " but WITHOUT ANY WARRANTY; without even the implied warranty of"
             " MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"
             " GNU Lesser General Public License for more details. ",
             " You should have received a copy of the GNU Lesser General"
             " Public License along with µGrid; see the file COPYING."
             " If not, write to the Free Software Foundation, Inc., 59"
             " Temple Place - Suite 330, Boston, MA 02111-1307, USA. ",
             " Additional permission under GNU GPL version 3 section 7 ",
             " If you modify this Program, or any covered work, by linking or"
             " combining it with proprietary FFT implementations or numerical"
             " libraries, containing parts covered by the terms of those"
             " libraries\' licenses, the licensors of this Program grant you"
             " additional permission to convey the resulting work. "]

py_lic_paras = ["µGrid is free software; you can redistribute it and/or\n"
                "modify it under the terms of the GNU Lesser General Public"
                " License as\npublished by the Free Software Foundation,"
                " either version 3, or (at\nyour option) any later version.",
                "µGrid is distributed in the hope that it will be useful,"
                " but\nWITHOUT ANY WARRANTY; without even the implied warranty"
                " of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See"
                " the GNU\nLesser General Public License for more details.",
                "You should have received a copy of the GNU Lesser General"
                " Public License\nalong with µGrid; see the file COPYING."
                " If not, write to the\nFree Software Foundation, Inc.,"
                " 59 Temple Place - Suite 330,\nBoston, MA 02111-1307, USA.",
                'Additional permission under GNU GPL version 3 section 7',
                "If you modify this Program, or any covered work, by linking"
                " or combining it\nwith proprietary FFT implementations or"
                " numerical libraries, containing parts\ncovered by the terms"
                " of those libraries' licenses, the licensors of this\nProgram"
                " grant you additional permission to convey the resulting"
                " work.\n"]

test_case = unittest.TestCase('__init__')


class CheckMuFFTHeaderFiles():

    def test_muGrid_header_files(self, muGrid_sources):
        msg_listh = ""
        msg_listh = lic_test.header_license_test(muGrid_sources, lic_paras)
        return msg_listh


class CheckMuFFTSourceFiles():

    def test_muGrid_source_files(self, muGrid_sources):
        msg_listc = ""
        msg_listc = lic_test.source_license_test(muGrid_sources, lic_paras)
        return msg_listc


class CheckMuFFTPythonFiles():

    def test_muGrid_python_files(self, muGrid_sources):
        msg_listp = ""
        msg_listp = lic_test.python_license_test(muGrid_sources, py_lic_paras)
        return msg_listp


def main():
    msg_list = ""
    muGrid_sources = []
    muGrid_sources = lic_test.arg_parser.parse_args(sys.argv[1:])

    header_test_case = CheckMuFFTHeaderFiles
    msg_list = msg_list + header_test_case.test_muGrid_header_files(
        header_test_case,
        muGrid_sources)

    source_test_case = CheckMuFFTSourceFiles
    msg_list = msg_list + source_test_case.test_muGrid_source_files(
        source_test_case,
        muGrid_sources)

    python_test_case = CheckMuFFTPythonFiles
    msg_list = msg_list + python_test_case.test_muGrid_python_files(
        python_test_case,
        muGrid_sources)

    test_case.assertEqual(len(msg_list), 0, msg_list)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file   Parallel.py

@author Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date   21 Mar 2018

@brief  muGrid Communicator object

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import _muGrid

try:
    from mpi4py import MPI
except ImportError:
    MPI = None


def Communicator(communicator=None):
    """
    Factory function for the communicator class.

    Parameters
    ----------
    communicator: mpi4py or muGrid communicator object
        The bare MPI communicator. (Default: _muGrid.Communicator())
    """
    # If the communicator is None, we return a communicator that contains just
    # the present process.
    if communicator is None:
        communicator = _muGrid.Communicator()

    # If the communicator is already an instance if _muGrid.Communicator, just
    # return that communicator.
    if isinstance(communicator, _muGrid.Communicator):
        return communicator

    # Now we need to do some magic. See if the communicator that was passed
    # conforms with the mpi4py interface, i.e. it has a method 'Get_size'.
    # The present magic enables using either mpi4py or stub implementations
    # of the same interface.
    if hasattr(communicator, 'Get_size'):
        # If the size of the communicator group is 1, just return a
        # communicator that contains just the present process.
        if communicator.Get_size() == 1:
            return _muGrid.Communicator()
        # Otherwise, check if muFFT does actually have MPI support. If yes
        # we assume that the communicator is an mpi4py communicator.
        elif _muGrid.Communicator.has_mpi:
            if not MPI:
                raise RuntimeError('muGrid was compiled with MPI support but '
                                   'mpi4py could not be loaded.')
            return _muGrid.Communicator(MPI._handleof(communicator))
        else:
            raise RuntimeError('muGrid was compiled without MPI support.')
    else:
        raise RuntimeError("The communicator does not have a 'Get_size' "
                           "method. muFFT only supports communicators that "
                           "conform to the mpi4py interface.")
